
local bigTable = {} 
-- This is the table that will hold a table for each boss,
-- which will hold a table for each event
-- which will hold a table for each raider that should get a message and those messages.
-- The boss table will also contain an init function (combat start) and onevent function.
-- Example: 
-- bigTable['Lei Shen'].init = SomeInitFunction
-- bigTable['Lei Shen'].onevent = SomeEventHandler
-- bigTable['Lei Shen'].onemote = SomeOtherEventHandler
-- bigTable['Lei Shen'].msgs['Transition 1 start']['Moujin'] = ['Go to red!']

-- Each onevent function is of the form func(spellName, event).
-- Each onemote function is of the form func(sender, msg).

-- Init functions take no parameters.\

-- The saved variable g_savedMessages has the same structure as bigTable but contains only msgs

local THIS_BOSS = nil             -- This will hold the current boss.

local lstRaids = {}         
-- This will be a list of the raids, for the dropdwons.
-- The keys will be just numbers, and each value will contain:
-- idx: its own index (the same number as above)
-- str: string identifier (e.g. 'ThroneOfThunder')
-- display: display string (e.g. 'Throne of Thunder')
-- bosses: list of boss names

local lstBossId = {}
-- This damn thing is to help me recognize the current boss fight.
-- For example, the "Council of Elders" fight doesn't actually have any mob named "Council of Elders",
-- so I need to take the mobs that do exist and see what fight they belong to...
-- Data will look like this: lstBossId['Frost King Malakk'] = 'Council of Elders'


-- Global references as well:
Raidbot_bigTable = bigTable
Raidbot_lstRaids = lstRaids
Raidbot_lstBossId = lstBossId

local CUSTOM_MSG_COUNT = 5
local CUSTOM_MSG = 'Custom message '
local EVERYONE = 'Everyone'

------------------------------------------------------------------
-- Messages and stuff

local HANDSHAKE     = 1     -- Check if someone has the add-on
local HANDSHAKE_ACK = 2     -- 
local MESSAGE       = 3     -- Send a message (instead of a whisper)
local PREVIEW       = 4     -- Send a message in preview (again, instead of a whisper)
local PREVIEW_CLEAR = 5     -- This is sent before a preview, so the recipient knows to clear his preview frame
local PREFIX = {
    [HANDSHAKE]     = 'RB_HANDSHAKE',
    [HANDSHAKE_ACK] = 'RB_HANDSHAKE_ACK',
    [MESSAGE]       = 'RB_MESSAGE',
    [PREVIEW]       = 'RB_PREVIEW',
    [PREVIEW_CLEAR] = 'RB_PREVIEW_CLEAR',
}
for k,v in pairs(PREFIX) do
    RegisterAddonMessagePrefix (v)
end

------------------------------------------------------------------

local function HandleIncomingMessage (prefix, message, channel, sender)
    if prefix==PREFIX[HANDSHAKE] then
        SendAddonMessage (PREFIX[HANDSHAKE_ACK], '', 'WHISPER', sender)
    elseif prefix==PREFIX[HANDSHAKE_ACK] then
        if frmRaidbot.raiders[sender] then
            frmRaidbot.raiders[sender].hasRaidbot = true
        end
    elseif prefix==PREFIX[MESSAGE] then
        PlaySoundFile ('Interface\\Addons\\Raidbot\\s.mp3', 'master')
        frmRaidbotDisplay:DisplayMessage (message)
    elseif prefix==PREFIX[PREVIEW] then
        frmRaidbotPreview:AddNote (message)
        frmRaidbotPreview:Show()
    elseif prefix==PREFIX[PREVIEW_CLEAR] then
        frmRaidbotPreview:ClearNotes ()
    end
    
end
------------------------------------------------------------------
------------------------------------------------------------------

local SIZEX = 160
local SIZEY_TITLE = 18
local SIZEY_NOTE = 22
local MAX_NOTES = 1

local SIZEY_RAIDER_FRAME = SIZEY_TITLE + (SIZEY_NOTE * MAX_NOTES) + 12
local RAIDERS_PER_LINE = 5

local ROLE_ICON_COORDS = {
    ['TANK'] = {0.0, 0.31, 0.32, 0.62},
    ['HEALER'] = {0.31, 0.62, 0.0, 0.31},
    ['DAMAGER'] = {0.31, 0.62, 0.32, 0.62},
}

local function RaidTargetIcon(n)
    if type(n)=='number' and (n>=1 and n<=8) then 
        return 'Interface\\TARGETINGFRAME\\UI-RaidTargetingIcon_'..tostring(n)
    else
        return ''
    end
end

------------------------------------------------------------------
------------------------------------------------------------------
-- Dialogs

local STR_CLEAR_CONFIRMATION = "Clear all messages for %s?"

StaticPopupDialogs["RAIDBOT_CLEAR"] = {
    --text = "Clear all messages for this ?",
    button1 = "Yes",
    button2 = "No",
    --[[OnAccept = function()
        frmRaidbot:ClearBoss ()
    end,]]
    timeout = 0,
    whileDead = true,
    hideOnEscape = true,
    preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
}

StaticPopupDialogs["RAIDBOT_CLEAREVENT"] = {
    --text = "Clear all messages for this ?",
    button1 = "Yes",
    button2 = "No",
    OnAccept = function()
        frmRaidbot:ClearEvent ()
    end,
    timeout = 0,
    whileDead = true,
    hideOnEscape = true,
    preferredIndex = 3,  -- avoid some UI taint, see http://www.wowace.com/announcements/how-to-avoid-some-ui-taint/
}

------------------------------------------------------------------

-- I have no idea why I have to use forward declarations... but it just doesn't work otherwise
local curMenuTargetPlayer   -- This variable will be set whenever the context menu is displayed for one of the raiders
local tblMenuRaiderFrame    -- The actual context menu

local function UpdateRaidTargetSubmenu (newMark)
    if tblMenuRaiderFrame==nil then return end
    local mark
    if newMark then
        mark = newMark
    else
        mark = GetRaidTargetIndex (curMenuTargetPlayer)
    end
    if mark==nil or mark==0 then 
        mark=9      -- 'None' is number 9 on the menu
    end
    for i,v in ipairs(tblMenuRaiderFrame[2].menuList) do
        v.checked = (i==mark)
    end
end

------------------------------------------------------------------

local function MarkRaider (self, arg1, arg2)
    SetRaidTarget (curMenuTargetPlayer, arg1)
    UpdateRaidTargetSubmenu(arg1)
end

------------------------------------------------------------------

local function RaidMarkerMenuEntry (n, txt)
    return {notCheckable = false, icon = RaidTargetIcon(n), func = MarkRaider, arg1 = n, text = txt}
end

------------------------------------------------------------------

tblMenuRaiderFrame = {
    { 
        text = "Options", 
        isTitle = true,
        notCheckable = true,
    },
    --[[{ 
        text = "Clear All", 
        notCheckable = true,
        func = function (self) 
            -- arg1 is the calling raider frame
            ClearRaiderNotes (curMenuTargetPlayer)
        end 
    },]]
    { 
        text = "Target Marker", 
        notCheckable = true,
        nested = true,
        hasArrow = true,
        menuList = {
            RaidMarkerMenuEntry (1, "Star"),
            RaidMarkerMenuEntry (2, "Circle"),
            RaidMarkerMenuEntry (3, "Diamond"),
            RaidMarkerMenuEntry (4, "Triangle"),
            RaidMarkerMenuEntry (5, "Moon"),
            RaidMarkerMenuEntry (6, "Square"),
            RaidMarkerMenuEntry (7, "Cross"),
            RaidMarkerMenuEntry (8, "Skull"),
            RaidMarkerMenuEntry (0, "None"),
        } 
    },
    { 
        text = "Cancel", 
        notCheckable = true,
    },
}
local cmenuRaiderFrame = CreateFrame("Frame", "menu", UIParent, "UIDropDownMenuTemplate")

------------------------------------------------------------------
            
local function DenullifyStr (s)
    return s or ''
end

------------------------------------------------------------------

local function FastTooltip (txt)
    GameTooltip:SetOwner (frmRaidbot, "ANCHOR_CURSOR")
    GameTooltip:AddLine (txt, 1.0, 1.0, 0.8)
    GameTooltip:Show ()
end

------------------------------------------------------------------

local function IsLeader (unit)
    -- Returns true if unit is group leader or assistant
    return unit~=nil and (UnitIsGroupLeader(unit) or UnitIsGroupAssistant(unit) )
end
------------------------------------------------------------------

local function CreateRaidbotBar ()
    local TITLE_W = 152
    local TITLE_H = 22
    local BUTTON_W = TITLE_W-20
    local BUTTON_H = 28
    local f = CreateFrame ("Frame", 'frmRaidbotBar', UIParent)
    f:SetSize (BUTTON_W, TITLE_H + BUTTON_H * CUSTOM_MSG_COUNT)
    f:SetPoint ("TOPRIGHT", frmRaidbot, "TOPLEFT")
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 12,
    })

    -- Frame
    f.frm = CreateFrame ("Frame", nil, f)
    f.frm:SetSize (BUTTON_W, BUTTON_H * CUSTOM_MSG_COUNT)
    f.frm:SetAllPoints (f)
    f.frm:SetFrameStrata("HIGH")    
    f.frm:SetBackdrop ({
            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 12,
    })

    -- Title
    f.title = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.title:SetPoint ("TOP", f, "TOP", 0, -4)
    f.title:SetSize (TITLE_W, TITLE_H)
    f.title:SetText('Raidbot')
    
    -- X button
    f.btnX = CreateFrame ("Button", nil, f, "UIPanelCloseButton")
    f.btnX:SetPoint("TOPRIGHT", f, "TOPRIGHT")
    f.btnX:SetSize(24, 24)


    -- Console button
    f.btns = {}
    for i=1,CUSTOM_MSG_COUNT do
        f.btns[i] = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
        f.btns[i].idx = i
        f.btns[i]:SetSize (BUTTON_W, BUTTON_H)
        f.btns[i]:SetPoint ("TOP", f, "TOP", 0, -TITLE_H - (i-1)*BUTTON_H)
        f.btns[i]:SetText (CUSTOM_MSG..i)
        f.btns[i]:SetScript ("OnClick", function (self,...)
            Raidbot_SendEventMessages (CUSTOM_MSG..self.idx)
        end)
    end
        
    -- Make it movable
    f:SetMovable (true)
    f:RegisterForDrag ("LeftButton")
    f:SetUserPlaced (true)
    f:SetClampedToScreen (true)
    f:EnableMouse (true)
    f:SetScript("OnMouseDown", function(self, button)
        self:StartMoving ()
    end)
    f:SetScript("OnMouseUp", function(self, button)
        self:StopMovingOrSizing ()
    end)
    
    -- Functions!
    function f:Update ()
        local count = 0
        for i=1,CUSTOM_MSG_COUNT do
            self.btns[i]:Disable ()
            local msgs = Raidbot_GetMessages(nil, CUSTOM_MSG..i)
            for k,v in pairs(msgs) do
                count = count + 1
                self.btns[i]:Enable ()
            end
        end
        
        if count > 0 then
            self:Show()
        else
            self:Hide()
        end
    end
    
    f:Hide()
    
    return f
end

------------------------------------------------------------------

local function ShowSmallButton (btn, frame, offsetx)
    btn.caller = frame
    btn:SetParent (frame)
    btn:SetPoint ("RIGHT", frame, "RIGHT", -offsetx, 0)
    btn:Show()
end

local function ShowSmallButtons (self)
    ShowSmallButton (frmRmBtnEdit, self, 18)
    ShowSmallButton (frmRmBtnClear, self, 2)
end

local function HideSmallButtons ()
    frmRmBtnEdit:Hide()
    frmRmBtnClear:Hide()
end

local function ShowNoteGlow (self)
    frmRmNoteGlow:SetPoint ("CENTER", self, "CENTER")
    frmRmNoteGlow:Show()
end

local function HideNoteGlow ()
    frmRmNoteGlow:Hide()
end

local function ShowNoteGlow2 (self)
    frmRmNoteGlow2:SetPoint ("CENTER", self, "CENTER")
    frmRmNoteGlow2:Show()
end

local function HideNoteGlow2 ()
    frmRmNoteGlow2:Hide()
end

------------------------------------------------------------------

local function _EditNote (self)
    if (not Raidbot_GetMessages()) or (not Raidbot_AnEventIsSelected()) then return end
    
    frmRaidbotInput:SetPoint ("CENTER", self, "CENTER")
    frmRaidbotInput.caller = self
    frmRaidbotInput.target = self:GetParent().name
    local txt = self:GetText ()
    frmRaidbotInput.txt:SetText (DenullifyStr(txt))
    frmRaidbotInput.txt:HighlightText()
    frmRaidbotInput:SetFrameStrata ('TOOLTIP')
    frmRaidbotInput:Show ()
    HideSmallButtons ()
end

------------------------------------------------------------------

local function _ClearNote (self)
    self:SetText ('')
    Raidbot_GetMessages()[self:GetParent().name] = nil
end

------------------------------------------------------------------

local function CreateFramedFontString (parent, name, w, h, txt)
    local f = CreateFrame ("Frame", name, parent)
    f:SetSize (w, h)
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 8,
    })

    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("TOPLEFT", f, "TOPLEFT", 4, 0)
    f.lbl:SetPoint ("BOTTOMRIGHT", f, "BOTTOMRIGHT", -4, 0)
    f.lbl:SetText (txt)
    
    return f
end

------------------------------------------------------------------
-- Raid leader console! Oh boy that's a long one

-- Drag & Drop...

local dragSrc
local dropDst
    
do
    local f = CreateFramedFontString (UIParent, 'frmRaidbot_drag', SIZEX, SIZEY_NOTE, '')
    f.lbl:SetTextColor (1.0,1.0,1.0)
    f.lbl:SetJustifyH ("LEFT")    
    
    function f:SetText (text)
        return self.lbl:SetText (text)
    end   
    function f:GetText ()
        return self.lbl:GetText ()
    end   

    f:SetFrameStrata ("TOOLTIP")
    f:SetMovable (true)
    f:SetAlpha (0.5)
    f:Hide()
end

------------------------------------------------------------------

local function StartDraggingNote (src)
    local txt = src:GetText()
    if txt=='' then return end
    
    frmRaidbot_drag:SetText (txt)
    frmRaidbot_drag:Show()
    frmRaidbot_drag:SetAllPoints (src)
    frmRaidbot_drag:StartMoving()
    ShowNoteGlow2 (src)
end

------------------------------------------------------------------

local function GetDropTarget ()
    -- Returns the name of the raider under the cursor
    for k,v in pairs(frmRaidbot.raiders) do
        if v:IsMouseOver (4, -26, 0, 0) then
            return k
        end
    end
end

------------------------------------------------------------------

local function DropNote ()
    local dst = GetDropTarget()
    if not dst then return end
    
    Raidbot_GetMessages()[dst] = frmRaidbot_drag:GetText()
    Raidbot_GetSavedMessages()[dst] = frmRaidbot_drag:GetText()
    frmRaidbot:LoadEventMessages()
    frmRaidbot.raiders[dst].noteFrames[1]:GetScript("OnEnter")()
end

------------------------------------------------------------------

local function StopDraggingNote ()
    DropNote ()
    dragSrc = nil
    dropDst = nil
    frmRaidbot_drag:StopMovingOrSizing()
    frmRaidbot_drag:Hide()
    HideNoteGlow2 ()
end

------------------------------------------------------------------

local function CreateRaiderFrameNote (parent, i)
    local ret = CreateFramedFontString (parent, nil, SIZEX, SIZEY_NOTE, '')
    ret:SetPoint ("TOPLEFT", parent, "BOTTOMLEFT", 0, -(i-1)*SIZEY_NOTE)
    ret.lbl:SetTextColor (1.0,1.0,1.0)
    ret.lbl:SetJustifyH ("LEFT")
    
    function ret:SetText (text)
        return self.lbl:SetText (text)
    end   
    function ret:GetText ()
        return self.lbl:GetText ()
    end   
    
    ret.EditNote = _EditNote
    ret.ClearNote = _ClearNote
    
    ret.idx = i
    ret:SetScript ("OnEnter", function (self,...)
        ShowSmallButtons (self)
        ShowNoteGlow (self)
        dropDst = self
    end)
    ret:SetScript ("OnLeave", function (self,...)
        HideSmallButtons ()
        HideNoteGlow ()
        dropDst = nil
    end)
    ret:SetScript ("OnMouseDown", function (self,...)
        StartDraggingNote (self)
    end)
    ret:SetScript ("OnMouseUp", function (self,...)
        HideNoteGlow ()
        StopDraggingNote (self)
        frmRaidbot.defaultPositions = false
    end)
    
    return ret
end

------------------------------------------------------------------

local function CreateRaiderFrame (name)
    if name==nil then return end
    
    local f = CreateFrame ("Frame", nil, frmRaidbot)
    f.name = name
    f:SetSize (SIZEX, SIZEY_TITLE)
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 4,
    })
    
    -- Player name
    local _,class = UnitClass(name)
    f.color = RAID_CLASS_COLORS[class]
    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("TOP", f, "TOP")
    f.lbl:SetSize (SIZEX, SIZEY_TITLE)
    f.lbl:SetText (name)
    f.lbl:SetTextColor (f.color.r, f.color.g, f.color.b)
    
    -- Highlight
    local margin = 1
    f.highlight = f:CreateTexture ('Texture', 'HIGH')
    f.highlight:SetPoint ("TOPLEFT",f, "TOPLEFT", margin, -margin)
    f.highlight:SetPoint ("BOTTOMRIGHT",f, "BOTTOMRIGHT", -margin, margin)
    f.highlight:SetTexture ("Interface\\Addons\\Raidbot\\glow")
    f.highlight:SetVertexColor (f.color.r, f.color.g, f.color.b,0.1)
    
    f:SetScript ("OnEnter", function (self, ...)
        f.highlight:SetAlpha (0.35)
    end)
    f:SetScript ("OnLeave", function (self, ...)
        f.highlight:SetAlpha (0.10)
    end)

    -- Role
    f.texRole = f:CreateTexture ('Texture', "HIGH")
    f.texRole:SetPoint ('TOPLEFT', f, 'TOPLEFT', 2, -1)
    f.texRole:SetSize (SIZEY_TITLE-2, SIZEY_TITLE-2)
    
    f.UpdateRole = function (self)
        f.role = UnitGroupRolesAssigned(self.name)
        if f.role~='NONE' then
            f.texRole:SetTexture ("Interface\\LFGFRAME\\UI-LFG-ICON-PORTRAITROLES")
            f.texRole:SetTexCoord (unpack(ROLE_ICON_COORDS[f.role]))
        else
            f.texRole:SetTexture ("")
        end
    end
    f:UpdateRole()
    
    -- Raid target
    f.texRaidTarget = f:CreateTexture ('Texture', "HIGH")
    f.texRaidTarget:SetPoint ('TOPRIGHT', f, 'TOPRIGHT', -2, -1)
    f.texRaidTarget:SetSize (SIZEY_TITLE-2, SIZEY_TITLE-2)
    
    f.UpdateRaidTarget = function (self)
        local raidTarget = GetRaidTargetIndex(self.name)
        if raidTarget~=nil then
            f.texRaidTarget:SetTexture (RaidTargetIcon(raidTarget))
        else
            f.texRaidTarget:SetTexture ("")
        end
    end
    f:UpdateRaidTarget()
    
    -- Notes
    f.notes = {}
    f.noteFrames = {}
    for i=1,MAX_NOTES do
        local tmp = CreateRaiderFrameNote (f, i)
        table.insert (f.noteFrames, tmp)
    end
    
    function f:SetText (text,idx)
        idx = idx or 1
        self.noteFrames[idx]:SetText (text)
    end
    function f:GetText (idx)
        idx = idx or 1
        return self.noteFrames[idx]:GetText ()
    end
    
    -- Make movable
    f:SetMovable (true)
    f:RegisterForDrag ("LeftButton")
    f:SetClampedToScreen (true)
    f:EnableMouse (true)
    f:SetScript("OnMouseDown", function (self, button)
        if button=='LeftButton' then 
            self:StartMoving () 
        elseif button=='RightButton' then
            HideSmallButtons ()
            curMenuTargetPlayer = self.name     -- Set the name of this raider in a global
            UpdateRaidTargetSubmenu ()
            EasyMenu (tblMenuRaiderFrame, cmenuRaiderFrame, "cursor", 0 , 20, "MENU");
        end
    end)
    f:SetScript("OnMouseUp", f.StopMovingOrSizing)    
    
    -- Same functions as in the normal Raidbot frame, kinda...
    f.Update = function (self)
        local s
        for i=1,MAX_NOTES do
            if self.notes[i] then
                s = self.notes[i]
            else
                s = ''
            end
            self.noteFrames[i].lbl:SetText (s)
        end
    end
    
    -- Send a handshake message, to see if this raider has Raidbot installed
    SendAddonMessage (PREFIX[HANDSHAKE], '', 'WHISPER', name)
        
    return f
end

------------------------------------------------------------------

local function CreateLoaderFrame ()
    if frmRaidbotLoader then return end
    
    local SLIDER_WIDTH = 19
    local TITLE_H = 46
    local BASE_H = 460
    local W, H = SIZEX + SLIDER_WIDTH + 16, BASE_H + TITLE_H

    local f = CreateFrame ("Frame", 'frmRaidbotLoader', UIParent)
    f:SetSize (W, H)
    f:SetPoint ("TOPLEFT", frmRaidbot, "TOPRIGHT")
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 12,
    })
    
    -- Title
    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("TOPLEFT", f, "TOPLEFT", 4, -6)
    f.lbl:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -4, -6)
    f.lbl:SetText ('Raidbot Loader')
    
    -- X button
    f.btnx = CreateFrame ("Button", nil, f, "UIPanelCloseButton")
    f.btnx:SetPoint("TOPRIGHT",f, "TOPRIGHT")
    f.btnx:SetSize(24,24)
    f.btnx:SetScript ("OnClick", function (self,...)
        frmRaidbotLoader:Hide ()
    end)
    
    -- Load all
    f.btnLoadAll = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnLoadAll:SetPoint("TOP",f, "TOP", 0 , -22)
    f.btnLoadAll:SetSize(W-12,22)
    f.btnLoadAll:SetText ('Load All')
    f.btnLoadAll:SetScript ("OnClick", function (self,...)
        for k,v in pairs(Raidbot_GetSavedMessages(frmRaidbot.selBoss, frmRaidbot.selEvent)) do
            if frmRaidbot.raiders[k] or k==EVERYONE then
                Raidbot_GetMessages()[k] = v
            end
            frmRaidbot:LoadEventMessages ()
        end
    end)
    
    do
        local f = CreateFrame ("ScrollFrame", nil, frmRaidbotLoader)
        frmRaidbotLoader.container = f
        f:SetSize (W, BASE_H-20)
        f:SetPoint ("TOP", frmRaidbotLoader, "TOP", 0, -TITLE_H)
        f:SetPoint ("BOTTOM", frmRaidbotLoader, "BOTTOM")
    
        -- Scrolling
        f:EnableMouseWheel (true)
        function f:scroll (delta)
            local val = self:GetVerticalScroll() - delta*20
            local _,maxval = self.scrl:GetMinMaxValues()
            
            if val < 0 then
                val = 0
            elseif val > maxval then
                val = maxval
            end
            self:SetVerticalScroll (val)
        end
        f:SetScript("OnMouseWheel", f.scroll)
        f:SetScript ("OnVerticalScroll", function(self, offset)
            self.scrl:SetValue (offset)
        end)
        
        -- The contents frame - this is the child frame
        local contents = CreateFrame ("Frame", nil, f)
        contents:SetSize (W - SLIDER_WIDTH, H)
        contents.sections = {}
        f:SetScrollChild (contents)
        f.contents = contents
        contents.msgs = {}
        
        -- The scroll bar
        f.scrl = CreateFrame ("Slider", nil, f)
        local s = f.scrl
        s:SetSize (SLIDER_WIDTH,0)
        s:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -4, -3)
        s:SetPoint ("BOTTOMRIGHT", f, "BOTTOMRIGHT", -4, 3)
        s:SetOrientation ("VERTICAL")
        s:SetBackdrop ({
            bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
            edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
            edgeSize = 2,
        })
        s:SetThumbTexture ("Interface\\Addons\\Raidbot\\thumb")
        s:SetMinMaxValues (0, max(0,contents:GetHeight()-f:GetHeight()))
        s:Enable ()
        s:SetValue (0)
        s:SetScript ("OnValueChanged", function (self, value)
            self:GetParent():SetVerticalScroll (value)
        end)
    end
    
    -- Loading saved messages
    function f:ShowSavedMessages ()
        local f = self.container.contents
        
        -- Clear current messages
        for i,v in ipairs(f.msgs) do
            v:Hide()
            v = nil
        end
        
        -- Get a nice list of new messages
        local msgs = {}
        local tmp = {}
        
        for k,v in pairs(Raidbot_GetSavedMessages(frmRaidbot.selBoss, frmRaidbot.selEvent)) do
            if not tmp[v] then
                tmp[v] = true
                table.insert (msgs, v)
            end
        end
        
        -- Show them
        for i,v in ipairs(msgs) do
            -- Create notes... but disable the stuff it gets automatically because the function was made for the raider frames
            local note = CreateRaiderFrameNote (f, 0)
            note:ClearAllPoints ()
            note:SetPoint ("TOP", f, "TOP", 0, -(i-1) * SIZEY_NOTE)
            note:SetScript ("OnEnter", function (self,...)
                ShowNoteGlow (self)
                dropDst = self
            end)
            
            note:SetText (v)
            f.msgs[i] = note
        end
        
        -- Adjust heights and whatnot
        local contents_height = #msgs * SIZEY_NOTE
        f:SetHeight (contents_height)
        self.container.scrl:SetMinMaxValues (0, max(0, contents_height-self.container:GetHeight()))
    end
    
    -- Make it movable
    f:SetMovable (true)
    f:RegisterForDrag ("LeftButton")
    f:SetUserPlaced (true)
    f:SetClampedToScreen (true)
    f:EnableMouse (true)
    f:SetScript("OnMouseDown", function(self, button)
        self:StartMoving ()
    end)
    f:SetScript("OnMouseUp", function(self, button)
        self:StopMovingOrSizing ()
    end)
    
    f:Hide()
end

------------------------------------------------------------------

local tblRoleComparison = {
    TANK=1, HEALER=2, DAMAGER=3, NONE=4,
    [1]=1, [2]=2, [3]=3, [4]=4,
}

local function CompareRaiderFrames (a,b)
    a.role = tblRoleComparison[a.role]
    b.role = tblRoleComparison[b.role]
    
    return (a.role < b.role) or (a.role==b.role and a.name<b.name)
end

------------------------------------------------------------------

local function ShowGeneralTooltip ()
    --[[GameTooltip:SetOwner (frmRaidbot, "ANCHOR_CURSOR")
    GameTooltip:AddLine ("Raidbot", 1.0, 0.9, 0.0)
    

    GameTooltip:Show ()]]
end

local function HideTooltip ()
    GameTooltip:Hide()
end

------------------------------------------------------------------

local function CreateRaidbotFrame ()
    local CONSOLE_H = 144
    local f = CreateFrame ("Frame", 'frmRaidbot', UIParent)
    f:SetSize (SIZEX*5, CONSOLE_H)
    f:SetPoint ("TOP", UIParent, "TOP", 0, -150)
    f:Hide()
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 16,
    })
    f.raiders = {}

    -- Title
    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("TOPLEFT", f, "TOPLEFT", 4, -6)
    f.lbl:SetPoint ("TOPRIGHT", f, "TOPRIGHT", -4, -6)
    f.lbl:SetText ('Raidbot')
    
    f.init = false
    f.raiders = {}
    f:SetScript ("OnShow", function(self)
        if not self.init then
            self.init = true
            if IsLeader('player') then
                for i=1,GetNumGroupMembers() do
                    local name,_,_,_,_,_,_,online = GetRaidRosterInfo(i)
                    if online then
                        --SendAddonMessage (PREFIX[SYNC], '', 'WHISPER', name)
                    end
                end
            end
        end
        self:Refresh ()
    end)
    
    -- Make it movable
    f:SetMovable (true)
    f:RegisterForDrag ("LeftButton")
    f:SetUserPlaced (true)
    f:SetClampedToScreen (true)
    f:EnableMouse (true)
    f:SetScript("OnMouseDown", f.StartMoving)
    f:SetScript("OnMouseUp", f.StopMovingOrSizing)
    
    -- X button
    f.btnX = CreateFrame ("Button", nil, f, "UIPanelCloseButton")
    f.btnX:SetPoint("TOPRIGHT", f, "TOPRIGHT")
    f.btnX:SetSize(SIZEY_TITLE*1.5,SIZEY_TITLE*1.5)

    -- ?
    f.help = CreateFrame ("Frame", nil, f)
    f.help:SetPoint ("TOPLEFT", f, "TOPLEFT", 2, 0)
    f.help:SetSize (18, SIZEY_NOTE-2)
    f.lblHelp = f.help:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lblHelp:SetAllPoints (f.help)
    f.lblHelp:SetTextColor (1.0,1.0,1.0)
    f.lblHelp:SetText ('?')
    f.help:SetScript ("OnEnter", ShowGeneralTooltip)
    f.help:SetScript ("OnLeave", HideTooltip)
    
    -- Refresh button
    --[[f.btnRefresh = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnRefresh:SetSize (75, SIZEY_TITLE-2)
    f.btnRefresh:SetPoint ("RIGHT", f.btnX, "LEFT", 0, 1)
    f.btnRefresh:SetText('Refresh')
    f.btnRefresh:SetScript("OnClick", function()
        f:Refresh()
    end)]]
    
    -- Default positions button
    f.btnDefault = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnDefault:SetSize (85, SIZEY_TITLE)
    f.btnDefault:SetPoint ("RIGHT", f.btnX, "LEFT", 0, 0)
    f.btnDefault:SetText('Default')
    f.btnDefault:SetScript("OnClick", function()
        f:SetDefaultPositions()
    end)
    
    -- Clear buttons
    f.btnClearBoss = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnClearBoss:SetSize (85, SIZEY_TITLE)
    f.btnClearBoss:SetPoint ("RIGHT", f.btnDefault, "LEFT")
    f.btnClearBoss:SetText('Clear Boss')
    f.btnClearBoss:SetScript("OnClick", function()
        if not frmRaidbot.selBoss then return end
        StaticPopupDialogs["RAIDBOT_CLEAR"]['text'] = STR_CLEAR_CONFIRMATION:format(frmRaidbot.selBoss)
        StaticPopupDialogs["RAIDBOT_CLEAR"]['OnAccept'] = frmRaidbot.ClearBoss
        StaticPopup_Show ("RAIDBOT_CLEAR")
    end)
    f.btnClearEvent = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnClearEvent:SetSize (85, SIZEY_TITLE)
    f.btnClearEvent:SetPoint ("RIGHT", f.btnClearBoss, "LEFT")
    f.btnClearEvent:SetText('Clear Event')
    f.btnClearEvent:SetScript("OnClick", function()
        if not frmRaidbot.selEvent then return end
        StaticPopupDialogs["RAIDBOT_CLEAR"]['text'] = STR_CLEAR_CONFIRMATION:format(frmRaidbot.selEvent)
        StaticPopupDialogs["RAIDBOT_CLEAR"]['OnAccept'] = frmRaidbot.ClearEvent
        StaticPopup_Show ("RAIDBOT_CLEAR")
    end)
    
    -- Layer 2: the dropdown boxes!
    local LAYER_2_Y = -59
    
    -- Raid list
    function f:DeselectEvent ()
        UIDropDownMenu_SetSelectedID (Raidbot_ddEvents, 0)
        UIDropDownMenu_SetText (Raidbot_ddEvents, '')
        self.selEvent = nil
        frmRaidbot:HideRaiderFrames()
    end
    function f:DeselectBoss ()
        UIDropDownMenu_SetSelectedID (Raidbot_ddBosses, 0)
        UIDropDownMenu_SetText (Raidbot_ddBosses, '')
        UIDropDownMenu_DisableDropDown (Raidbot_ddEvents)
        self:DeselectEvent()
        self.selBoss = nil
        frmRaidbot:HideRaiderFrames()
        frmRaidbot.btnPreview:Disable()
    end
    f.ddRaids = CreateFrame("Button", "Raidbot_ddRaids", f, "UIDropDownMenuTemplate")
    f.ddRaids:SetPoint("TOPLEFT", f, "TOPLEFT", 10, LAYER_2_Y)
    UIDropDownMenu_Initialize (f.ddRaids, function (self, level)
        local info = UIDropDownMenu_CreateInfo()
        for i,v in ipairs(lstRaids) do
            info = UIDropDownMenu_CreateInfo()
            info.text = v.display
            info.value = v.idx
            info.func = function (self)
                UIDropDownMenu_SetSelectedID(f.ddRaids, self:GetID())
                Raidbot_ddBosses:init (self.value)
                frmRaidbot:DeselectBoss()
            end
            UIDropDownMenu_AddButton(info, level)
        end
    end)
    UIDropDownMenu_SetWidth (f.ddRaids, 160);
    UIDropDownMenu_SetButtonWidth (f.ddRaids, 160)
    UIDropDownMenu_JustifyText (f.ddRaids, "LEFT")
    
    -- Boss list
    f.ddBosses = CreateFrame("Button", "Raidbot_ddBosses", f, "UIDropDownMenuTemplate")
    f.ddBosses:SetPoint("TOPLEFT", f, "TOPLEFT", 240, LAYER_2_Y)
    function f.ddBosses:init (idxRaid)
        UIDropDownMenu_Initialize (self, function (self, level)
            local info = UIDropDownMenu_CreateInfo()
            for i,v in ipairs(lstRaids[idxRaid].bosses) do
                info = UIDropDownMenu_CreateInfo()
                info.text = v
                info.value = v
                info.func = function (self)
                    UIDropDownMenu_SetSelectedID(f.ddBosses, self:GetID())
                    frmRaidbot.selBoss = self.value
                    frmRaidbot:DeselectEvent()
                    frmRaidbot.btnPreview:Enable()
                    Raidbot_ddEvents:init (self.value)          -- this call changes self, and I have no idea why                    
                end
                UIDropDownMenu_AddButton(info, level)
            end
        end)
        UIDropDownMenu_EnableDropDown (Raidbot_ddBosses)
    end
    UIDropDownMenu_SetWidth (f.ddBosses, 160);
    UIDropDownMenu_SetButtonWidth (f.ddBosses, 160)
    UIDropDownMenu_JustifyText (f.ddBosses, "LEFT")
    
    -- Event list
    f.ddEvents = CreateFrame("Button", "Raidbot_ddEvents", f, "UIDropDownMenuTemplate")
    f.ddEvents:SetPoint("TOPLEFT", f, "TOPLEFT", 470, LAYER_2_Y)
    function f.ddEvents:init (boss)
        UIDropDownMenu_Initialize (self, function (self, level)
            local info = UIDropDownMenu_CreateInfo()
            for i,v in ipairs(bigTable[boss].events) do
                info = UIDropDownMenu_CreateInfo()
                info.text = v
                info.value = v
                info.func = function (self)
                    UIDropDownMenu_SetSelectedID(f.ddEvents, self:GetID())
                    frmRaidbot.selEvent = self.value
                    frmRaidbot:LoadEventMessages ()
                    frmRaidbot:ShowRaiderFrames()
                end
                UIDropDownMenu_AddButton(info, level)
            end
        end)
        UIDropDownMenu_EnableDropDown (Raidbot_ddEvents)
    end
    UIDropDownMenu_SetWidth (f.ddEvents, 160);
    UIDropDownMenu_SetButtonWidth (f.ddEvents, 160)
    UIDropDownMenu_JustifyText (f.ddEvents, "LEFT")
    
    UIDropDownMenu_DisableDropDown (Raidbot_ddBosses)
    UIDropDownMenu_DisableDropDown (Raidbot_ddEvents)
    
    -- Labels for these dropdowns
    local LABEL_W = 145
    local LABEL_H = 22
    
    f.lblRaids = f:CreateFontString (nil, "BACKGROUND", "GameFontNormal")
    f.lblRaids:SetSize (LABEL_W, LABEL_H)
    f.lblRaids:SetPoint ("BOTTOMLEFT", f.ddRaids, "TOPLEFT")
    f.lblRaids:SetText ('Raid')
    
    f.lblBosses = f:CreateFontString (nil, "BACKGROUND", "GameFontNormal")
    f.lblBosses:SetSize (LABEL_W, LABEL_H)
    f.lblBosses:SetPoint ("BOTTOMLEFT", f.ddBosses, "TOPLEFT")
    f.lblBosses:SetText ('Boss')

    f.lblEvents = f:CreateFontString (nil, "BACKGROUND", "GameFontNormal")
    f.lblEvents:SetSize (LABEL_W, LABEL_H)
    f.lblEvents:SetPoint ("BOTTOMLEFT", f.ddEvents, "TOPLEFT")
    f.lblEvents:SetText ('Event')

    -- Preview button
    f.btnPreview = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnPreview:SetSize (85, 22)
    f.btnPreview:SetPoint("TOPLEFT", f, "TOPLEFT", 693, LAYER_2_Y-2)
    f.btnPreview:SetText('Preview')
    f.btnPreview.tooltipText = 'Send all the messages associated with this boss right now,\nto let people know the plan in advance'
    f.btnPreview:SetScript("OnClick", function(self)
        Raidbot_PreviewBossMessages (frmRaidbot.selBoss)
    end)
    

    -- Layer 3: raid warning
    local LAYER_3_Y = LAYER_2_Y - 46
        
    f.txtEveryone = CreateFrame ("EditBox", nil, f)
    f.txtEveryone:SetPoint ("TOPLEFT", f, "TOPLEFT", 180, LAYER_3_Y)
    f.txtEveryone:SetSize (225, 24)
    f.txtEveryone:SetTextColor (1.0,1.0,1.0,1.0)
    f.txtEveryone:SetFont ('Fonts\\ARIALN.TTF', 14, "")
    f.txtEveryone:SetTextInsets (7, 7, 2, 2)
    f.txtEveryone:SetFont ("Fonts\\FRIZQT__.TTF", 12)
    f.txtEveryone:SetAutoFocus (false)
    f.txtEveryone:SetBackdrop ({
        bgFile = "Interface\\Addons\\Raidbot\\black",
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
        edgeSize = 12,
    })
    
    f.txtEveryone:SetScript ("OnEditFocusGained", function(self)
        self.oldVal = self:GetText()
        self:HighlightText()
    end)
    f.txtEveryone:SetScript ("OnEnterPressed", function(self)
        if self:GetText()~='' then
            Raidbot_GetMessages()[EVERYONE] = self:GetText()
            Raidbot_GetSavedMessages()[EVERYONE] = self:GetText()
        else
            Raidbot_GetMessages()[EVERYONE] = nil
        end        
        self:ClearFocus()
    end)
    f.txtEveryone:SetScript ("OnEscapePressed", function(self)
        self:SetText(self.oldVal)
        self:ClearFocus ()
    end)

    f.lblEveryone = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lblEveryone:SetPoint ("TOPLEFT", f, "TOPLEFT", 30, LAYER_3_Y)
    f.lblEveryone:SetSize (140, 20)
    f.lblEveryone:SetJustifyH ("LEFT")
    f.lblEveryone:SetText ("Message to everyone:")
    
    -- Loader button
    f.btnLoader = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnLoader:SetSize (85, 22)
    f.btnLoader:SetPoint("TOPLEFT", f, "TOPLEFT", 693, LAYER_3_Y-2)
    f.btnLoader:SetText('Load')
    f.btnLoader.tooltipText = 'Load previously used messages for this event'
    f.btnLoader:SetScript("OnClick", function(self)
        frmRaidbotLoader:Show()
    end)

    -- Copy/ Paste
    f.btnPaste = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnPaste:SetSize (85, 22)
    f.btnPaste:SetPoint("RIGHT", f.btnLoader, "LEFT", -20, 0)
    f.btnPaste:SetText('Paste')
    f.btnPaste:Disable ()
    f.btnPaste:SetScript("OnClick", function(self)
        if (not frmRaidbot.copyBoss) or (not frmRaidbot.copyEvent) then return end
        for k,v in pairs(Raidbot_GetMessages (frmRaidbot.copyBoss, frmRaidbot.copyEvent)) do
            Raidbot_GetMessages()[k] = v
        end
        frmRaidbot:LoadEventMessages()
    end)
    
    f.btnCopy = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btnCopy:SetSize (85, 22)
    f.btnCopy:SetPoint("RIGHT", f.btnPaste, "LEFT", -20, 0)
    f.btnCopy:SetText('Copy')
    f.btnCopy.tooltipText = 'Copy all the messages from this event, so you can paste them to another'
    f.btnCopy:SetScript("OnClick", function(self)
        local isEmpty = true
        for k,v in pairs(Raidbot_GetMessages()) do
            isEmpty = false
            break
        end
        
        if isEmpty then
            FastTooltip ('No messages to copy!')
        else
            frmRaidbot.btnPaste:Enable ()
            frmRaidbot.btnPaste.tooltipText = format('Paste all messages from %s (%s)',frmRaidbot.selEvent, frmRaidbot.selBoss)
            frmRaidbot.copyBoss = frmRaidbot.selBoss
            frmRaidbot.copyEvent = frmRaidbot.selEvent
        end
    end)
    
    
    -- Now about those raider frames...
    f.CalculateRaiderFrameLocation = function (self, idx)
        local x = (idx-1) % RAIDERS_PER_LINE
        local y = floor((idx-1) / RAIDERS_PER_LINE)
        return x*SIZEX, y*(SIZEY_RAIDER_FRAME+2)
    end
    
    -- Functions!
    function f:SetDefaultPositions ()
        local tmp = {}
        for k,v in pairs(self.raiders) do
            table.insert (tmp, {name=k, role=v.role})
        end
        sort (tmp, CompareRaiderFrames)
        
        for idx,raider in ipairs(tmp) do
            local x,y = f:CalculateRaiderFrameLocation (idx)
            f.raiders[raider.name]:SetPoint ("TOPLEFT", f, "BOTTOMLEFT", x, -y)
        end
    end
    function f:ClearEvent ()
        for k,v in pairs(frmRaidbot.raiders) do
            Raidbot_GetMessages()[k] = nil
        end
        frmRaidbot:LoadEventMessages()
    end
    function f:ClearBoss ()
        for k,v in pairs(bigTable[frmRaidbot.selBoss].msgs) do
            for name,frm in pairs(frmRaidbot.raiders) do
                Raidbot_GetMessages(frmRaidbot.selBoss, k)[name] = nil
            end
        end
        frmRaidbot:LoadEventMessages()
    end
    
    f.defaultPositions = true
    
    function f:HideRaiderFrames ()
        for k,v in pairs(self.raiders) do v:Hide() end
        self.lblEveryone:Hide()
        self.txtEveryone:Hide()
        frmRaidbotInput:Hide()
        self.btnLoader:Disable()                            -- Disable loader button
    end
    function f:ShowRaiderFrames ()
        for k,v in pairs(self.raiders) do v:Show() end
        self.lblEveryone:Show()
        self.txtEveryone:Show()
        self.btnLoader:Enable()                             -- Enable loader button
        frmRaidbotLoader:ShowSavedMessages()                -- Load saved messages for this event
    end
    
    -- Refreshing!
    -- Events that cause refreshing may happen frequently, so I'm buffering them
    f.Refresh = function (f)
        f.NeedsRefresh = true
        f.elapsed = 0
    end
    f.ForceRefresh = function (f)
        if not f.raiders then return end
    
        -- Get the list of everyone in the raid
        local raid = {}
        for i=1,GetNumGroupMembers() do
            local tmp = GetRaidRosterInfo(i)
            if tmp then raid[tmp] = true end
        end
        
        -- Get rid of the frames of people who are no longer in the raid
        -- Those who still are... at least update their roles
        for k,v in pairs(f.raiders) do
            if raid[k]==nil then
                v:Hide()
                f.raiders[k] = nil
            else
                v:UpdateRole()
                v:UpdateRaidTarget()
            end
        end
        
        -- Now add frames for people who joined the raid
        for i=1,GetNumGroupMembers() do
            local name,_,_,_,_,_,_,online = GetRaidRosterInfo(i)
            if name==nil then
                f.NeedsRefresh = true
                return
            end
            if f.raiders[name]==nil then
                local x,y = f:CalculateRaiderFrameLocation (i)
                --while i>GetNumGroupMembers() do i=i-GetNumGroupMembers() end
                local tmp = CreateRaiderFrame (name)
                if tmp then
                    tmp:SetPoint ("TOPLEFT", f, "BOTTOMLEFT", x, -y)
                    f.raiders[name] = tmp
                end
            end
        end
        
        if f.defaultPositions then
            f:SetDefaultPositions()
        end
        
        if f.selEvent then
            f:ShowRaiderFrames()
        else
            f:HideRaiderFrames()
        end
    end
    f:SetScript ("OnUpdate", function (f, elapsed)
        f.elapsed = f.elapsed + elapsed
        if f.elapsed > 0.5 then
            if f.NeedsRefresh then
                f.NeedsRefresh = false
                f:ForceRefresh ()
            end
            f.elapsed = 0
        end
        
    end)
    f:ForceRefresh ()
    
    -- Actually displaying the raider messages
    function f:LoadEventMessages ()
        for k,v in pairs(self.raiders) do
            v:SetText (Raidbot_GetMessages()[k])
        end
        self.txtEveryone:SetText (DenullifyStr(Raidbot_GetMessages()[EVERYONE]))
        frmRaidbotBar:Update()
        frmRaidbotInput:Hide()        
    end
    
    f:RegisterEvent ("GROUP_ROSTER_UPDATE")
    f:RegisterEvent ("RAID_TARGET_UPDATE")
    f:RegisterEvent ("CHAT_MSG_ADDON")
    f:SetScript ("OnEvent", function(self, event, prefix, message, channel, sender, ...)
        if event=='CHAT_MSG_ADDON' then
            HandleIncomingMessage (prefix, message, channel, sender)
        elseif event=="GROUP_ROSTER_UPDATE" or event=="RAID_TARGET_UPDATE" then
            self:Refresh()
        end
    end)
    
    f:DeselectBoss()
    
    if g_savedMessages then
    
    else
        g_savedMessages = {}
    end
end


------------------------------------------------------------------

local function CreateSmallButton (name, img)
    local f = CreateFrame ("Frame", name, frmRaidbot)
    f:Hide ()
    f:SetSize (SIZEY_NOTE-6,SIZEY_NOTE-6)
    f:SetFrameStrata ('TOOLTIP')
    f:Raise()
    
    f.btn = CreateFrame ("Button", nil, f, "UIPanelButtonTemplate")
    f.btn:SetAllPoints (f)
    f.btn:SetNormalTexture (img)
    f.btn:SetPushedTexture (img)
    
    return f
end

------------------------------------------------------------------

local function CreateEditButton ()
    -- Well, it's one global button that moves wherever it's needed...
    local f = CreateSmallButton ('frmRmBtnEdit', 'Interface\\BUTTONS\\UI-GuildButton-PublicNote-Up')
    
    -- Whoever calls this must set frmRmBtnEdit.caller
    f.btn:SetScript("OnClick", function()
        frmRmBtnEdit.caller:EditNote ()
    end)
    
    f.btn:SetScript("OnEnter", function (self,...)
        frmRmBtnEdit.caller:GetScript ("OnEnter") (frmRmBtnEdit.caller)
    end)
    f.btn:SetScript("OnLeave", function (self,...)
        frmRmBtnClear.caller:GetScript ("OnLeave") (frmRmBtnClear.caller)
    end)
end

------------------------------------------------------------------

local function CreateClearButton ()
    -- Exactly the same as the edit button
    local f = CreateSmallButton ('frmRmBtnClear', 'Interface\\Addons\\Raidbot\\x')
        
    -- Whoever calls this must set frmRmBtnClear.caller
    f.btn:SetScript("OnClick", function()
        frmRmBtnClear.caller:ClearNote ()
    end)
    
    f.btn:SetScript("OnEnter", function (self,...)
        frmRmBtnClear.caller:GetScript ("OnEnter") (frmRmBtnClear.caller)
    end)
    f.btn:SetScript("OnLeave", function (self,...)
        frmRmBtnClear.caller:GetScript ("OnLeave") (frmRmBtnClear.caller)
    end)
end

------------------------------------------------------------------

local function _CreateNoteGlow (name)
    local f = CreateFrame ("Frame", name, frmRaidbot)
    f:Hide()
    f:SetSize (SIZEX, SIZEY_NOTE)
    f:SetFrameStrata ('TOOLTIP')
    f:Lower ()
    
    f.tex = f:CreateTexture ('Texture', 'HIGH')
    f.tex:SetAllPoints (f)
    f.tex:SetTexture ("Interface\\Addons\\Raidbot\\glow")
    f.tex:SetBlendMode ("ADD")
    f.tex:SetVertexColor (0.3, 0.5, 0.8, 0.25)
    
    return f
end

local function CreateNoteGlow ()
    return _CreateNoteGlow('frmRmNoteGlow')
end

local function CreateNoteGlow2 ()
    local ret = _CreateNoteGlow('frmRmNoteGlow2')
    ret.tex:SetAlpha (0.4)
    return ret
end

------------------------------------------------------------------

local function CreateNoteInputFrame ()
    local f = CreateFrame ("Frame", 'frmRaidbotInput', frmRaidbot)
    f:SetSize (SIZEX+6,SIZEY_NOTE+6)
    f:SetPoint ("CENTER", UIParent, "CENTER")
    f:SetBackdrop ({
        bgFile = "Interface\\Addons\\Raidbot\\black",
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
        edgeSize = 14,
    })
    f:SetFrameStrata ('TOOLTIP')
    f:Hide()
    
    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("CENTER", f, "CENTER")
    f.lbl:SetSize (SIZEX-8, SIZEY_NOTE-2)
    f.lbl:SetTextColor (1.0,1.0,1.0)
    f.lbl:SetJustifyH ("LEFT")
    
    f.txt = CreateFrame ("EditBox", nil, f)
    f.txt:SetAllPoints(f)
    f.txt:SetTextColor (1.0,1.0,1.0,1.0)
    f.txt:SetFont ('Fonts\\ARIALN.TTF', 14, "")
    f.txt:SetTextInsets (7, 7, 2, 2)
    f.txt:SetFont ("Fonts\\FRIZQT__.TTF", 12)
    --f.txt:SetFocus()

    f.txt:SetScript ("OnEnterPressed", function(self)
        local p = self:GetParent()
        if self:GetText()~='' then
            Raidbot_GetMessages()[p.target] = self:GetText()
            Raidbot_GetSavedMessages()[p.target] = self:GetText()
        else
            Raidbot_GetMessages()[p.target] = nil
        end
        --p.caller:SetText (self:GetText())
        frmRaidbot:LoadEventMessages()
        p:Hide()
    end)
    f.txt:SetScript ("OnEscapePressed", function(self)
        self:GetParent():Hide()
        self:SetText('')
    end)
end

------------------------------------------------------------------

local function CreateDisplayFrame ()
    local f = CreateFrame ("Frame", 'frmRaidbotDisplay', UIParent)
    f:SetSize (630,120)
    f:SetPoint ("CENTER", UIParent, "CENTER", 0, -140)
    f:SetFrameStrata ('TOOLTIP')
    
    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetAllPoints (f)
    f.lbl:SetTextColor (1.0,0.8,0.0)
    f.lbl:SetFont ('Fonts\\FRIZQT__.ttf', 72, "THICKOUTLINE")
    f.lbl:SetJustifyH ("CENTER")
    f.lbl:SetShadowColor (0,0,0, 0.8)
    f.lbl:SetShadowOffset (2, -2)
    
    -- Movable
    f:SetMovable (true)
    f:RegisterForDrag ("LeftButton")
    f:SetUserPlaced (true)
    f:SetClampedToScreen (true)
            
    -- Animations
    f.animGroup = f:CreateAnimationGroup()
    
    f.animAlpha = f.animGroup:CreateAnimation("Alpha")
    f.animAlpha:SetChange (-1)
    f.animAlpha:SetDuration (2.5)
    f.animAlpha:SetStartDelay (4.5)
    f.animAlpha:SetSmoothing("IN")
    f.animAlpha:SetOrder (1)

    f.animGroup:SetScript ("OnFinished", function (self, requested)
        frmRaidbotDisplay:Hide()
    end)
    
    -- Some functions
    function f:DisplayMessage (msg)
        self.lbl:SetText (msg)
        self:Show ()
        self.animGroup:Stop ()
        self.animGroup:Play ()
    end
    
    function f:MakeMovable ()
        self:SetBackdrop ({
            bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",
        })
        self.lbl:SetText ('Move Me!')
        
        self:EnableMouse (true)
        self:SetScript("OnMouseDown", function(self, button)
            self:StartMoving ()
        end)
        
        self:SetScript("OnMouseUp", function(self, button)
            self:StopMovingOrSizing ()
            self:SetBackdrop ({bgFile = ''})
            self:EnableMouse (false)
            self:SetScript("OnMouseDown", nil)
            self:DisplayMessage ('This is the new position')
        end)
        self:Show()
    end
    
    
    return f
end

------------------------------------------------------------------

local function CreateNote (self)
    local f = CreateFrame ("Frame", nil, self)
    f:SetSize (self:GetWidth(),SIZEY_NOTE)
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
        edgeSize = 4,
    })

    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetPoint ("CENTER", f, "CENTER")
    f.lbl:SetSize (self:GetWidth() - 14,SIZEY_NOTE)
    f.lbl:SetTextColor (1.0,1.0,1.0)
    f.lbl:SetJustifyH ("LEFT")
    
    return f
end
------------------------------------------------------------------

local function CreatePreviewFrame ()
    local f = CreateFrame ("Frame", 'frmRaidbotPreview', UIParent)
    f:SetSize (SIZEX*1.4, SIZEY_TITLE)
    f:SetPoint ("CENTER", UIParent, "CENTER")
    f:SetBackdrop ({
        bgFile = "Interface\\DialogFrame\\UI-DialogBox-Background",  
        edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
        edgeSize = 8,
    })
    
    -- Title
    f.lbl = f:CreateFontString(nil, "BACKGROUND", "GameFontNormal")
    f.lbl:SetAllPoints (f)
    f.lbl:SetJustifyH ("CENTER")
    f.lbl:SetText ('Raidbot')
    
    -- X button
    f.btnX = CreateFrame ("EditBox", nil, f)
    f.btnX:SetPoint ("TOPRIGHT", f, "TOPRIGHT")
    f.btnX:SetPoint ("BOTTOMLEFT", f, "BOTTOMRIGHT", -13, 0)
    f.btnX:SetTextColor (f.lbl:GetTextColor())
    f.btnX:SetTextInsets (0, 0, 2, 2)
    f.btnX:SetFont ("Fonts\\FRIZQT__.TTF", 10)
    f.btnX:SetAutoFocus (false)
    f.btnX:Disable ()
    f.btnX:SetText ('X')
    f.btnX:SetScript ("OnMouseDown", function (self, button)
        frmRaidbotPreview:Hide()
    end)
    
    -- Notes...
    f.noteFrames = {[0]=f}
    f.noteCount = 0
    f.CreateNote = CreateNote
    function f:AddNote (txt)
        self.noteCount = self.noteCount + 1
        if self.noteCount > #self.noteFrames then
            table.insert (self.noteFrames, self:CreateNote ())
            self.noteFrames[#self.noteFrames]:SetPoint ("TOP", self.noteFrames[#self.noteFrames-1], "BOTTOM")
        end
        self.noteFrames[self.noteCount]:Show()
        self.noteFrames[self.noteCount].lbl:SetText (txt)
    end
    function f:ClearNotes ()
        self.noteCount = 0
        for i,v in ipairs(self.noteFrames) do
            v:Hide()
        end
    end
    
    -- Make it movable
    f:SetMovable (true)
    f:RegisterForDrag ("LeftButton")
    f:SetUserPlaced (true)
    f:SetClampedToScreen (true)
    f:EnableMouse (true)
    f:SetScript("OnMouseDown", function(self, button)
        self:StartMoving ()
    end)
    f:SetScript("OnMouseUp", function(self, button)
        self:StopMovingOrSizing ()
    end)
    
    f:Hide()
    
    return f
end

------------------------------------------------------------------

local function UpdateParents ()
    -- What a bloody annoying system
    frmRaidbotInput:SetParent (frmRaidbot)
    frmRmNoteGlow:SetParent (frmRaidbot)
    frmRmNoteGlow2:SetParent (frmRaidbot)
    frmRmBtnEdit:SetParent (frmRaidbot)
    frmRmBtnClear:SetParent (frmRaidbot)
end

------------------------------------------------------------------
------------------------------------------------------------------
-- Gotta love Blizzard's sense of humor...
-- Some code has to run right when loading (creating some frames),
-- while other code is absolutely forbidden to (stuff like querying
-- your raid roster will just return nils)
--
-- So here, most (not all) of the stuff is set to run a few seconds after login.
-- Ha!

local f = CreateFrame ("Frame", nil, UIParent)
f:RegisterEvent ("PLAYER_ENTERING_WORLD")
f:SetScript ("OnEvent", function (self,...)
    f.ready = true
end)
f.elapsed = 0
f:SetScript ("OnUpdate", function (self, elapsed,...)
    if f.ready then
        f.elapsed = f.elapsed + elapsed
    end
    if f.ready and f.elapsed > 3 and frmRaidbot==nil and GetRaidRosterInfo(1)~=nil then
        -- It's really difficult to make sure this runs only after everything is loaded :s
        f.ready = false
        CreateNoteInputFrame ()
        CreateEditButton ()
        CreateClearButton ()
        CreateNoteGlow ()
        CreateNoteGlow2 ()
        CreateRaidbotFrame ()
        CreateRaidbotBar ()
        CreateLoaderFrame ()
        
        UpdateParents ()
    end
end)

-- Frames that don't really care about any of the above problems, but I want to remember their positions
CreateDisplayFrame ()
CreatePreviewFrame ()

------------------------------------------------------------------

local function AnyoneIsInCombat ()
    if GetNumGroupMembers()==0 then
        return UnitAffectingCombat('player')
    end
    
    local groupType
    if IsInRaid() then
        groupType = 'raid'
    else
        groupType = 'party'
        if UnitAffectingCombat('player') then
            return true
        end
    end
    
    for i=1,GetNumGroupMembers() do
        if UnitAffectingCombat(groupType..tostring(i)) then
            return true
        end            
    end
    return false
end

------------------------------------------------------------------

Raidbot_LoadData_ThroneOfThunder (bigTable, lstRaids, lstBossId)
-- Add custom messages to each boss
for idx,name in ipairs(lstRaids[#lstRaids].bosses) do
    for i=1,CUSTOM_MSG_COUNT do
        bigTable[name].msgs[CUSTOM_MSG..tostring(i)] = {}
        table.insert (bigTable[name].events, 'Custom message '..tostring(i))
    end
end



------------------------------------------------------------------
-- Init stuff for each fight

Raidbot_counter = {}
local function StartCombat (boss)
    if not bigTable[boss] then
        boss = lstBossId[boss]
    end
    
    if (not boss) or (not bigTable[boss]) then return end
    
    Raidbot_counter = {}
    THIS_BOSS = boss
    bigTable[THIS_BOSS].init ()
end

local function EndCombat ()
    THIS_BOSS = nil
end

local function IsWaitingToStartCombat()
    return THIS_BOSS==nil
end

------------------------------------------------------------------

local function OnEvent (spellName, event)
    if not THIS_BOSS then return end
    bigTable[THIS_BOSS].onevent (spellName, event)
end

local function OnEmote (sender, msg)
    if not THIS_BOSS then return end
    bigTable[THIS_BOSS].onemote (sender, msg)
end
------------------------------------------------------------------

local function SendSingleMessage (name, msg, preview)
    if name==EVERYONE then
        SendChatMessage(msg, "RAID_WARNING")
    else
        if frmRaidbot.raiders[name].hasRaidbot then
            SendAddonMessage (PREFIX[MESSAGE], msg, 'WHISPER', name)
        else
            SendChatMessage(msg, "WHISPER", nil, name)
        end
    end
end

------------------------------------------------------------------

function Raidbot_SendEventMessages (event)
    --print ('Event: '..event)
    if not THIS_BOSS then return end
    for name,msg in pairs(bigTable[THIS_BOSS].msgs[event]) do
        if name==EVERYONE then
            SendChatMessage(msg, "RAID_WARNING")
        else
            if frmRaidbot.raiders[name].hasRaidbot then
                SendAddonMessage (PREFIX[MESSAGE], msg, 'WHISPER', name)
            else
                SendChatMessage(msg, "WHISPER", nil, name)
            end
        end
    end
end

------------------------------------------------------------------

local function _EventNameForPreview (event)
    if event:match(CUSTOM_MSG) then
        return '*'
    else
        return event..' - '
    end
end

------------------------------------------------------------------

local function SendPreviewMessages (boss)
    for i,event in ipairs(bigTable[boss].events) do
        for name, msg in pairs(bigTable[boss].msgs[event]) do
            local msg = _EventNameForPreview(event)..' '..msg
            if name==EVERYONE then
                SendAddonMessage (PREFIX[PREVIEW], msg, 'RAID')
                for k,v in pairs(frmRaidbot.raiders) do
                    if not v.hasRaidbot then
                        SendChatMessage(msg, "WHISPER", nil, k)
                    end
                end
            else
                if frmRaidbot.raiders[name].hasRaidbot then
                    SendAddonMessage (PREFIX[PREVIEW], msg, 'WHISPER', name)
                else
                    SendChatMessage(msg, "WHISPER", nil, name)
                end
            end
        end
    end
end

------------------------------------------------------------------

function Raidbot_PreviewBossMessages (boss)
    -- When sending several addon messages, I can't assume they'll arrive in the correct order.
    -- So I have to send the clear message first, then delay before sending the preview messages.
    if not boss then return end
    SendAddonMessage (PREFIX[PREVIEW_CLEAR], '', 'RAID')
    frmRaidbotWatcher.previewBoss = boss
end
local PreviewBossMessages = Raidbot_PreviewBossMessages

------------------------------------------------------------------

do
    local f = CreateFrame ("Frame", 'frmRaidbotWatcher', UIParent)
    f:RegisterEvent ("COMBAT_LOG_EVENT_UNFILTERED")
    f:RegisterEvent ("PLAYER_REGEN_DISABLED")
    f:RegisterEvent ("PLAYER_REGEN_ENABLED")
    f:RegisterEvent ("CHAT_MSG_RAID_BOSS_EMOTE")
    f:SetScript ("OnEvent", function  (bla, e, timeStamp, event, hideCaster, srcGuid, srcName, srcFlags, srcRaidFlags, dstGuid, dstName, dstFlags, dstRaidFlags, ...)
        if e=='PLAYER_REGEN_ENABLED' then
            --
        
        elseif e=='PLAYER_REGEN_DISABLED' then
            --
        
        elseif e=='CHAT_MSG_RAID_BOSS_EMOTE' then
            local message, sender, language, channelString, target, flags, unknown, channelNumber, channelName = timeStamp, event, hideCaster, srcGuid, srcName, srcFlags, srcRaidFlags, dstGuid, dstName
            OnEmote (sender, message)
            
        elseif e=='COMBAT_LOG_EVENT_UNFILTERED' then
            if AnyoneIsInCombat() and IsWaitingToStartCombat() and dstName then
                StartCombat (dstName)
                if THIS_BOSS then 
                    --print ('Combat started: '..THIS_BOSS)
                end
            end
            
            if not THIS_BOSS then return end
            if not AnyoneIsInCombat() then
                --print (format('Combat ended: %s', THIS_BOSS))
                EndCombat ()
                return
            end
            
            if event:match('SPELL') then
                local details = {...}            
                OnEvent (details[2], event)
            end
            
        end
    end)
    
    f.elapsed = 0
    f:SetScript ("OnUpdate", function (self, elapsed)
        if self.previewBoss then 
            self.elapsed = self.elapsed + elapsed
            
            if self.elapsed > 1 then
                SendPreviewMessages (self.previewBoss)
                self.elapsed = 0
                self.previewBoss = nil            
            end
        end
    end)
end

------------------------------------------------------------------
-- Util

function Raidbot_GetMessages (boss, event)
    boss = boss or frmRaidbot.selBoss
    event = event or frmRaidbot.selEvent
    
    if boss==nil or event==nil or bigTable[boss]==nil or bigTable[boss].msgs[event]==nil then
        return {}
    else
        return bigTable[boss].msgs[event]
    end
end

------------------------------------------------------------------

function Raidbot_GetSavedMessages (boss, event)
    boss = boss or frmRaidbot.selBoss
    event = event or frmRaidbot.selEvent
        
    if boss==nil or event==nil then
        return {}
    else
        g_savedMessages[boss] = g_savedMessages[boss] or {msgs={}}
        g_savedMessages[boss].msgs[event] = g_savedMessages[boss].msgs[event] or {}
        return g_savedMessages[boss].msgs[event]
    end
end

------------------------------------------------------------------

function Raidbot_AnEventIsSelected ()
    return frmRaidbot.selBoss and frmRaidbot.selEvent
end

------------------------------------------------------------------


-- Slash commands
SLASH_RAIDBOT1 = "/raidbot"
SlashCmdList["RAIDBOT"] = function (txt)
    if #txt==0 then
        if frmRaidbot then
            frmRaidbot:Show()
        else
            print ('But... but... you aren\'t in a raid group! :s')
        end
    elseif txt=='move' then
        frmRaidbotDisplay:MakeMovable ()
    else
        print ('Sorry, the only commands I know are "/raidbot" and "/raidbot move" :(')
    end
end
